/*
 * Copyright (C) 2006 - 2009
 * Thomas van Dijk
 * Jan-Pieter van den Heuvel
 * Wouter Slob
 * William van Hemert
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 
package libtw.ngraph;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;

import libtw.input.GraphInput.InputData;
import libtw.reductions.ReductionData;

public abstract class NGraph< D > implements Iterable<NVertex<D>>
{
	
	String comments = "";
	
	public abstract void addVertex( NVertex<D> v );
	public abstract void removeVertex( NVertex<D> v );
	
	public abstract NVertex<D> getVertex( int i );
	public abstract Iterator<NVertex<D>> getVertices();
	public abstract int getNumberOfVertices();
	
	public void addComment( String c )
	{
		comments = comments.concat( "\n" + c );
	}

	public String getComments()
	{
		return comments;
	}
	
	public Iterator<NVertex<D>> iterator()
	{
		return getVertices();
	}

	public void eliminate( NVertex<D> v ) {
		for( NVertex<D> n1 : v ) {
			for( NVertex<D> n2 : v ) {
				if( n1!=n2 ) n1.ensureNeighbor( n2 );
			}
		}
		removeVertex( v );
	}

	public ArrayList<NEdge<D>> eliminate2( NVertex<D> v ) {
		ArrayList<NEdge<D>> es = new ArrayList<NEdge<D>>();
		for( NVertex<D> n1 : v ) {
			for( NVertex<D> n2 : v ) {
				if( n1!=n2 && n1.ensureNeighbor( n2 ) ) {
					es.add( new NEdge<D>(n1,n2) );
				}
			}
		}
		removeVertex( v );
		return es;
	}
	public void deEliminate( NVertex<D> v, ArrayList<NEdge<D>> es ) {
		addVertex( v );
		for( NVertex<D> n : v ) {
			n.ensureNeighbor(v);
		}
		for( NEdge<D> e : es ) {
			e.a.removeNeighbor( e.b );
		}
	}
	
	public void ensureEdge( NVertex<D> a, NVertex<D> b ) {
		a.ensureNeighbor( b );
		b.ensureNeighbor( a );
	}
	public void addEdge( NVertex<D> a, NVertex<D> b ) {
		a.addNeighbor( b );
		b.addNeighbor( a );
	}
	
	public Iterable<NEdge<D>> edges() {
		final NGraph<D> g = this;
		return new Iterable<NEdge<D>>() {
			public Iterator<NEdge<D>> iterator() {
				HashSet<NEdge<D>> edges = new HashSet<NEdge<D>>();
				for( NVertex<D> v : g ) {
					for( NVertex<D> n : v ) {
						edges.add( new NEdge<D>(v,n) );
					}
				}
				return edges.iterator();
			}
			
		};
	}
	
	public int getNumberOfEdges() {
		int n = 0;
		for( NVertex<D> v : this ) {
			n += v.getNumberOfNeighbors();
		}
		return n/2;
	}
	
	public void contractEdge(NEdge<D> e) {
		contractEdge(e.a,e.b);
	}
	
	public void contractEdge(NVertex<D> a, NVertex<D> b) {
		NVertex<D> v1,v2;

		//Select the vertex with the highest degree
		if(a.getNumberOfNeighbors() < b.getNumberOfNeighbors()) {
			v1 = b;
			v2 = a;
		} else {
			v1 = a;
			v2 = b;
		}
		
		//Remove the edge from the edgelists of the vertex
		v1.removeNeighbor(v2);
		v2.removeNeighbor(v1);
		
		//Add vertices of v1 to the list and remove the old edge from those vertices
		for(NVertex<D> ed : v2) {
			
			ensureEdge(v1,ed);
		}
		
		//Remove the other vertex
		removeVertex(v2);
	}
	
	/**
	 * This function will test the vertex for the islet rule.
	 * 
	 * @author William van Hemert
	 * @param v the vertex under consideration
	 * @return true, if the islet rule can be applied on the vertex
	 */
	public boolean testIslet(NVertex<D> v) {
		// check if the degree is zero 
		return v.getNumberOfNeighbors() == 0;
	}
	
	/**
	 * This function will test the vertex for the twig rule.
	 * 
	 * @author William van Hemert
	 * @param v the vertex under consideration
	 * @return true, if the twig rule can be applied on the vertex
	 */
	public boolean testTwig(NVertex<D> v) {
		// check if the degree is one		
		return v.getNumberOfNeighbors() == 1;
	}

	/**
	 * This function will test the vertex for the series rule.
	 * 
	 * @author William van Hemert
	 * @param v the vertex under consideration
	 * @return true, if the series rule can be applied on the vertex
	 */
	public boolean testSeries(NVertex<D> v) {
		// check if the degree is two
		return v.getNumberOfNeighbors() == 2;
	}
	
	/**
	 * This function will test the vertex for the triangle rule.
	 * 
	 * @author William van Hemert
	 * @param v the vertex under consideration
	 * @return true, if the triangle rule can be applied on the vertex
	 */
	public boolean testTriangle(NVertex<D> v) {
		// the number of neighbors must be exactly three
		if (v.getNumberOfNeighbors() == 3) {
			// pick the iterator from the vertex
			Iterator<NVertex<D>> it = v.getNeighbors();
			// store the three neighbors in variables
			NVertex<D> v1 = it.next(); 
			NVertex<D> v2 = it.next();
			NVertex<D> v3 = it.next();
			
			// if one of the neighbors is connected to another vertex
			if (v1.isNeighbor(v2) || v1.isNeighbor(v3) || v2.isNeighbor(v3))
				return true;
		}
		
		// if degree is unequal to three or the three neighbors have no connection at all
		return false;
	}

	/**
	 * This function will test the vertex for the simplicial rule.
	 * 
	 * @author William van Hemert
	 * @param v the vertex under consideration
	 * @return true, if the simplicial rule can be applied on the vertex
	 */
	public boolean testSimplicial(NVertex<D> v) {

		// check each neighboring vertex
		for( NVertex<D> n : v )
			// if the degree is smaller than the degree from the vertex under consideration ...
			if (n.getNumberOfNeighbors() < v.getNumberOfNeighbors())
				// ... then it could not be a simplical vertex, so return false
				return false;

		// check each neighboring vertex
		for(NVertex<D> n1: v)
			// check each neighboring vertex once again
			for(NVertex<D> n2: v)
				// if the two neighbors are different ...
				if(n1.data != n2.data)
					// ... then they must be connected, if not ... 
					if(!n1.isNeighbor(n2))
						// ... then it could not be a simplical vertex, so return false
						return false;
		
		// if both checks pass, return true
		return true;
	}

	/**
	 * This function will do a comparison with the almost simplicial rule.
	 * 
	 * @author William van Hemert
	 * @param v the vertex under consideration
	 * @return true, if the almost simplicial rule can be applied on the vertex
	 */
	public boolean testAlmostSimplicial(NVertex<D> v) {
		// returns true if the returned value not is null
		return testAlmostSimplicialVertex(v) != null;
	}
	
	/**
	 * This function will test the vertex for the almost simplicial vertex rule.
	 * 
	 * @author William van Hemert
	 * @param v the vertex under consideration
	 * @return null if there is no almost simplicial neighbor, otherwise the neighbor is returned
	 */
	public NVertex<D> testAlmostSimplicialVertex(NVertex<D> v) {
		int c = 0;
		
		// check each neighbor
		for( NVertex<D> n : v )
			// if the degree is smaller than the degree of the vertex under consideration
			if (n.getNumberOfNeighbors() < v.getNumberOfNeighbors() - 1)
				// increase the counter
				c++;

		// if there are two vertices with a lower degree the vertex could not be a almost simplicial vertex
		if (c > 1)
			return null;
		
		// check each neighbor and take it as almost simplicial neighbor
		for(NVertex<D> asv: v) {
			// initialize asvb to true
			boolean asvb = true;
			int asvn = 0;

			// check each neighbor
			for(NVertex<D> n: v)
				// if the neighbor is unequal to the almost simplicial neighbor
				if (n.data != asv.data)
					// if they are neighbors ...
					if (n.isNeighbor(asv))
						// ... update the counter
						asvn++;
			
			// if the degree is equal to the number of neighbors
			if (asvn == v.getNumberOfNeighbors() -1)
				// then the almost simplicial neighbor is member of the clique, set asvb to false 
				asvb = false;
			
			if (asvb)
				// check whether or not the other neighbors are a clique, using two loops
				for(NVertex<D> n1: v) 
					for(NVertex<D> n2: v) 
						// if v1 and v2 are unequal and they are different from the almost simplicial neighbor ...  
						if((n1.data != n2.data) && (n1.data != asv.data) && (n2.data != asv.data))
							// ... and if they are not neighbors ...
							if(!n1.isNeighbor(n2))
								// ... they do not form a clique, set asvb to false
								asvb = false;

			// if asvb is still true
			if (asvb)
				// return the almost simplicial neighbor under consideration
				return asv;
		}
		
		return null;
	}
	
	/**
	 * This function will test the vertex for the buddy rule.
	 * 
	 * @author William van Hemert
	 * @param v1 the vertex under consideration
	 * @return the vertex that is the buddy of v1
	 */
	public NVertex<D> testBuddy(NVertex<D> v1) {
		// the number of neighbors must be exactly three 
		if (v1.getNumberOfNeighbors() == 3)	{
			// take a random neighbor
			NVertex<D> neighbour = v1.getNeighbors().next();
			// check each neighbor of that neighbor
			for( NVertex<D> v2 : neighbour ) {
				// if the neighbor differs from the vertex under consideration
				if (v2.data != v1.data) {
					// if the degree from that vertex is also three
					if (v2.getNumberOfNeighbors() == 3) {
						// create two ArrayLists to store the ID's
						ArrayList<Integer> v1Ids = new ArrayList<Integer>();
						ArrayList<Integer> v2Ids = new ArrayList<Integer>();
						
						// fill the first ArrayList with the neighboring ID's of v1
						for( NVertex<D> n1 : v1 ) {
							InputData input = (InputData) ((ReductionData<?>) n1.data).original.data; 
							v1Ids.add(input.id);
						}
						Collections.sort(v1Ids);

						// fill the second ArrayList with the neighboring ID's of v2
						for( NVertex<D> n2 : v2 ) {
							InputData input = (InputData) ((ReductionData<?>) n2.data).original.data;
							v2Ids.add(input.id);
						}
						Collections.sort(v2Ids);
	
						// compare the two sets, if they are equal
						if (v1Ids.containsAll(v2Ids))
							// return the neighbor of the neighbor of v1
							return v2;
					}
				}
			}
		}
		
		// return null if no buddy was found
		return null;
	}
	
	/**
	 * This function will test the vertex for the buddies rule.
	 * 
	 * @author William van Hemert
	 * @param v1 the vertex under consideration
	 * @return all vertices that share the same neighbors as v1, including v1
	 */
	public ArrayList<NVertex<D>> testBuddies(NVertex<D> v1) {
		ArrayList<NVertex<D>> buddies = new ArrayList<NVertex<D>>();
		ArrayList<Integer> v1Ids = new ArrayList<Integer>();

		// if the vertex has no neighbors
		if (v1.getNumberOfNeighbors() == 0)
			// return an empty list
			return new ArrayList<NVertex<D>>();
		
		// check each neighbor
		for( NVertex<D> n1 : v1 ) {
			// take the original data object from the vertex
			InputData input = (InputData) ((ReductionData<?>) n1.data).original.data;
			// add the ID to the set
			v1Ids.add(input.id);
		}
		// sort the set and add v1 to the buddies set
		Collections.sort(v1Ids);
		buddies.add(v1);
		
		// pick a random neighbor
		NVertex<D> neighbour = v1.getNeighbors().next();
		
		// check each neighbor of that neighbor 
		for( NVertex<D> v2 : neighbour ) {
			// if the neighbor differs from the vertex under consideration
			if (v2.data != v1.data) {
				// if the degree is equal to the vertex under consideration
				if (v2.getNumberOfNeighbors() == v1.getNumberOfNeighbors()) {
					// create a list to store the ID's of v2
					ArrayList<Integer> v2Ids = new ArrayList<Integer>();
					
					// fill the second ArrayList with the neighboring ID's of v2
					for( NVertex<D> n2 : v2 ) {
						InputData input = (InputData) ((ReductionData<?>) n2.data).original.data;
						v2Ids.add(input.id);
					}
					Collections.sort(v2Ids);

					// if the set is equal to the set of v1 ...
					if (v1Ids.containsAll(v2Ids))
						// ... add v2 to the buddies
						buddies.add(v2);
				}
			}
		}
		
		// return the set with buddies
		return buddies;
	}
	
	/**
	 * This function will do a comparison with the cube rule.
	 * 
	 * @author William van Hemert
	 * @param v the vertex under consideration
	 * @return true, if the almost cube rule can be applied on the vertex
	 */
	public boolean testCube(NVertex<D> v) {
		// returns true if exactly three neighbors are found for the cube
		return testCubeVertex(v).size() == 3;
	}
	
	/**
	 * This function will test the vertex for the cube vertex rule.
	 * 
	 * @author William van Hemert
	 * @param v the vertex under consideration
	 * @return returns the vertices that are member of the cube
	 */
	public HashSet<NVertex<D>> testCubeVertex(NVertex<D> v) {
		// the number of neighbors must be exactly three
		if (v.getNumberOfNeighbors() == 3) {
			HashSet<NVertex<D>> cube = new HashSet<NVertex<D>>();
			ArrayList<NVertex<D>> threeNeighbors = new ArrayList<NVertex<D>>();
			
			// check each neighbor
			for (NVertex<D> neighbor : v)
				// if the degree is three ...
				if (neighbor.getNumberOfNeighbors() == 3)
					// add it to the set
					threeNeighbors.add(neighbor);
					
			// if the vertex has less than three neighbors with degree three
			if (threeNeighbors.size() < 3)
				// return a empty vertex set
				return new HashSet<NVertex<D>>();

			// check each vertex in the threeNeighbors set
			for (NVertex<D> tv : threeNeighbors)
				// check each neighbor of the vertex 
				for (NVertex<D> tn : tv)
					// if the neighbor is not the original vertex
					if (tn.data != v.data)
						// and of the vertex is not contained in the set already
						if (!threeNeighbors.contains(tn))
							// add the vertex to the set
							cube.add(tn);
				
			// if the cube contains exactly three unique neighbors of neighbors
			if (cube.size() == 3)
				// return the cube
				return cube;
		} 
		
		// if no cube with three unique neighbors is found, return a empty set
		return new HashSet<NVertex<D>>();	
	}

	/**
	 * This function will test the vertex for the extended cube vertex rule.
	 * 
	 * @author William van Hemert
	 * @param v the vertex under consideration
	 * @return the set of neighboring vertices that created the extended cube, multiple sets could be returned
	 */
	public ArrayList<ArrayList<NVertex<D>>> testExtendedCubeVertex(NVertex<D> v) {
		// the degree of the vertex must be higher than or equal to three
		if (v.getNumberOfNeighbors() >= 3) {
			ArrayList<ArrayList<NVertex<D>>> extendedCubes = new ArrayList<ArrayList<NVertex<D>>>();
			ArrayList<NVertex<D>> threeNeighbors = new ArrayList<NVertex<D>>();
			HashSet<ArrayList<NVertex<D>>> commonNeighborPairs = new HashSet<ArrayList<NVertex<D>>>();
			
			// check each neighbor
			for (NVertex<D> neighbor : v)
				// if the degree is three ...
				if (neighbor.getNumberOfNeighbors() == 3)
					// add it to the set
					threeNeighbors.add(neighbor);
	
			// if the vertex has less than three neighbors with degree three
			if (threeNeighbors.size() < 3)
				// return a empty vertex set
				return new ArrayList<ArrayList<NVertex<D>>>();
			
			// check each vertex with a degree of three
			for (NVertex<D> v1 : threeNeighbors)
				// check each vertex with a degree of three
				for (NVertex<D> v2 : threeNeighbors)
					// if the two neighbors are different
					if (v1.data != v2.data)
						// find the common neighbor by looping over all combinations
						for (NVertex<D> n1 : v1)
							for (NVertex<D> n2 : v2)
								// if the common neighbor is found and the neighbor is not the current vertex
								if ((n1.data == n2.data) && (n1.data != v.data))
									// perform an extra check, otherwise a clique of three is recognized as extended cube
									if (!threeNeighbors.contains(n1)) {  
										ArrayList<NVertex<D>> commonNeighborPair = new ArrayList<NVertex<D>>();
										// read the original ID's from the two vertices
										int id1 = ((InputData) ((ReductionData<?>) v1.data).original.data).id;
										int id2 = ((InputData) ((ReductionData<?>) v2.data).original.data).id;
										
										// keep the ordering in the ID's
										// otherwise two items with swapped objects are added to commonNeighborPairs 
										if (id1 < id2) { 
											commonNeighborPair.add(v1);
											commonNeighborPair.add(v2);
										} else {
											commonNeighborPair.add(v2);
											commonNeighborPair.add(v1);
										}
										
										// add the common neighbor pair
										commonNeighborPairs.add(commonNeighborPair);
									}
	
			// if there are less than three neighbor pairs ...
			if (commonNeighborPairs.size() < 3)
				// ... no cube can be created, and an empty cube set is returned
				return new ArrayList<ArrayList<NVertex<D>>>();
			
			ArrayList<NVertex<D>> extendedCube = commonNeighborPairs.iterator().next();
			boolean added;
			
			// iterate over all pairs and try to find a loop with three pairs
			for (ArrayList<NVertex<D>> pair : commonNeighborPairs){
				// if pair is the first element of the commonNeighborPairs list, it is
				// already added to the extendedCube list before the for loop began
				if (!commonNeighborPairs.iterator().next().equals(pair)) {
					added = false;
					
					// in case the last element of extendedCube is the first member of pair
					if ((extendedCube.get(extendedCube.size()-1) == pair.get(0)) && (!added)) {
						// add the last vertex to the current pair
						extendedCube.add(pair.get(1));
						added = true;
					}
						
					// in case the last element of extendedCube is the last member of pair
					if ((extendedCube.get(extendedCube.size()-1) == pair.get(1)) && (!added)) {
						// add the first vertex to the current pair
						extendedCube.add(pair.get(0));
						added = true;
					}
					
					// in case the first element of extendedCube is the first member of pair
					if ((extendedCube.get(0) == pair.get(0)) && (!added)) {
						// create a new list
						ArrayList<NVertex<D>> extendedCubeTemp = (ArrayList<NVertex<D>>) extendedCube.clone();
						extendedCube.clear();
						// add the last element						
						extendedCube.add(pair.get(1));
						// add all other elements
						extendedCube.addAll(extendedCubeTemp);
						added = true;
					}
					
					// in case the first element of extendedCube is the last member of pair
					if ((extendedCube.get(0) == pair.get(1)) && (!added)) {
						// create a new list
						ArrayList<NVertex<D>> extendedCubeTemp = (ArrayList<NVertex<D>>) extendedCube.clone();
						extendedCube.clear();
						// add the first element
						extendedCube.add(pair.get(0));
						// add all other elements
						extendedCube.addAll(extendedCubeTemp);
						added = true;
					}
					
					// in case the list is a loop
					if (extendedCube.get(0) == extendedCube.get(extendedCube.size()-1) && extendedCube.size() == 4)  {
						// remove the last element from the cube
						extendedCube.remove(extendedCube.size()-1); // remove last element again
						// add the cube to the cube set set
						extendedCubes.add((ArrayList<NVertex<D>>) extendedCube.clone());
					}
				}
			}
			
			// return the list with cube sets
			return extendedCubes;
		} else 
			// return a new list
			return new ArrayList<ArrayList<NVertex<D>>>();
	}
	
	public abstract void setVertices( ArrayList<NVertex<D>> vs );
	
	public static interface Convertor< From, To > {
		public To convert( NVertex<From> v );
	}
	public <To> NGraph<To> copy( Convertor<D,To> c ) {
				
		NGraph<To> newG = new ListGraph<To>();
		
		HashMap<NVertex<D>, NVertex<To>> oldToNew = new HashMap<NVertex<D>, NVertex<To>>();
		
		for( NVertex<D> v : this ) {
			To newData = c.convert( v );
			NVertex<To> nv = v.newOfSameType( newData );
			newG.addVertex( nv );
			oldToNew.put( v, nv );
		}
		for( NVertex<D> v : this ) {
			for( NVertex<D> n : v ) {
				oldToNew.get(v).addNeighbor(oldToNew.get(n));
			}
		}
		
		return newG;
		
	}
	
	public NGraph<D> copy() {
		
		NGraph<D> newG = new ListGraph<D>();
		
		HashMap<NVertex<D>, NVertex<D>> oldToNew = new HashMap<NVertex<D>, NVertex<D>>();
		
		for( NVertex<D> v : this ) {
			NVertex<D> nv = v.newOfSameType( v.data );
			newG.addVertex( nv );
			oldToNew.put( v, nv );
		}
		for( NVertex<D> v : this ) {
			for( NVertex<D> n : v ) {
				oldToNew.get(v).addNeighbor(oldToNew.get(n));
			}
		}
		
		return newG;
		
	}

	
}
