/*
 * Copyright (C) 2013
 * 
 * Stefan Fafianie
 *
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * This package makes use of a selection of classes from the libtw library 
 * which is also published under the GNU Lesser General Public license.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this package; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 * 
 */

package order;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.List;

import libtw.input.GraphInput.InputData;

/* Introduce edge node with a single child. 
 * If one of the end points of the edge is not contained in the subset of this node 
 * then we simply return solutions from the child node in the same order.
 * In the case that we can introduce the edge we will return partial solutions
 * from the child node until it becomes cheaper to return partial solutions
 * which we have previously returned, but with the edge included. */
public class IntroduceEdgeNode<D extends InputData> extends Node<D> {

	Node<D> childNode;
	int vertexIDA, vertexIDB, weight;
	boolean addEdge;
	PartialSolution edgePrint;
	List<PartialSolution> solutions;
	boolean askNext, cleanup;
	
	public IntroduceEdgeNode(BitSet subset, int vertexIDA, int vertexIDB, int weight, OrderDP<D> dp){
		super(subset, dp);
		this.subset = subset;
		this.vertexIDA = vertexIDA;
		this.vertexIDB = vertexIDB;
		this.weight = weight;
		
		if (subset.get(vertexIDA) && subset.get(vertexIDB)){
			addEdge = true;
			edgePrint = PartialSolution.edgePrint(subset, vertexIDA, vertexIDB, weight);
			solutions = new ArrayList<PartialSolution>();
			askNext = true;
			cleanup = false;
		}
		else{
			addEdge = false;
			reduce = false;
		}
	}
	
	public PartialSolution applyNext() {
		if (addEdge){
			PartialSolution result, first, last;
			
			if (askNext && childNode.hasNext){
				PartialSolution next = childNode.reduceNext();
				if (next == null){
					cleanup = true;
				}
				else{
					solutions.add(next);
				}
				
				askNext = false;
			}
			
			if (solutions.isEmpty()){
				hasNext = false;
				return null;
			}
			first = solutions.get(0);
			last = solutions.get(solutions.size()-1);
			
			if (first.cost + edgePrint.cost < last.cost || cleanup ){ 			
				result = first.joinPartialSolution(edgePrint);
				dp.operations++;
				solutions.remove(0);
				if (solutions.isEmpty()){
					hasNext = false;
				}
			}
			else{
				result = last;
				askNext = true;
				if (!childNode.hasNext){
					cleanup = true;
				}
			}
			
			return result;
		}
		else{
			if (childNode.hasNext){
				return childNode.reduceNext();
			}
			else{
				hasNext = false;
				return null;
			}
		}
	}

	@Override
	public String toString() {
		return "edge " + vertexIDA + "-" + vertexIDB + " " + addEdge;
	}

}
