/*
 * Copyright (C) 2013
 * 
 * Stefan Fafianie
 *
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * This package makes use of a selection of classes from the libtw library 
 * which is also published under the GNU Lesser General Public license.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this package; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 * 
 */

package order;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.HashSet;
import java.util.List;

import libtw.input.GraphInput.InputData;

/* Node corresponding to a subset of vertices in a bag in the tree decomposition
 * Nodes keep some internal state to in order to return partial solutions in order of optimality
 * Dominated and linearly dependent partial solutions are eliminated
 * Nodes have a hasNext flag which can be checked to see if a child has more partial solutions to offer.
 * In some cases a request for a partial solution may return null (if the last partial solution is independent) however
 * Nodes must make sure that this scenario is dealt with correctly */
public abstract class Node<D extends InputData> {

	public boolean hasNext, reduce, checkUnique;
	BitSet subset;
	
	List<Integer> leading;
	List<BitSet> rows;
	HashSet<BitSet> unique;
	OrderDP<D> dp;
	int independent;
	int max;
	
	//Check if a partial solution is independent
	boolean independent(BitSet row, int size){
		
		//Dominated partial solutions are dependent
		if (unique.contains(row) || independent >= max){
			return false;
		}
		else{
			unique.add(row);
		}
		
		//Apply row operations to a copy, keep the result and return the original		
		BitSet newRow = (BitSet) row.clone();
			
		//Apply row operations from previous partial solutions
		int i = 0;
		while (i < rows.size() ){
			if (newRow.get(leading.get(i))){
				newRow.clear(leading.get(i));
				for (int j = leading.get(i)+1; j < size ; j++){
					if (rows.get(i).get(j )){
						newRow.flip(j);
					}
				}
			}
			
			i++;				
		}
		
		int k = 0;
		boolean found = false;
		while (!found && k < size){
			if (newRow.get(k)){
				found = true;
				leading.add(k);
				rows.add(newRow);
			}
			k++;
		}
		if (found){
			independent++;
		}
		return found;
	}
	
	public Node(BitSet subset, OrderDP<D> dp){
		checkUnique = true;
		reduce = true;
		hasNext = true;
		leading = new ArrayList<Integer>();
		rows = new ArrayList<BitSet>();
		unique = new HashSet<BitSet>();
		this.dp = dp;
		independent = 0;
		max = (int) Math.pow(2, subset.cardinality()-1);
	}
	
	public PartialSolution reduceNext(){
		
		boolean searching = true;
		PartialSolution p = applyNext();
		
		//Keep requesting partial solutions until we find one that is independent
		while (searching){
			
			if (p == null || p.size == 0 || !reduce || independent(p.row, p.size)){
				searching = false;
			}
			else{
				p = applyNext();
			}
			
		}
		
		
		
		if (independent >= max){
			hasNext = false;
		}
		return p;
		
		
		
		
		
		
		
		
	}
	
	public abstract PartialSolution applyNext();
	
	public abstract String toString();
}
