/*
 * Copyright (C) 2013
 * 
 * Stefan Fafianie
 *
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * This package makes use of a selection of classes from the libtw library 
 * which is also published under the GNU Lesser General Public license.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this package; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 * 
 */

package steiner;

import java.util.*;

import libtw.input.GraphInput.InputData;
import libtw.ngraph.NVertex;

/* Straightforward implementation of the rank-based approach to reduce the size
 * of tables computed during the classic DP for Steiner-tree.  */
public class RankBasedSteinerTreeDP <D extends InputData> extends SteinerTreeDP<D> {
	
	boolean always = false;
	
	@Override
	public HashMap<Set<NVertex<D>>,HashMap<Connectivity<D>,Integer>> calculateBag(NVertex<NTDNiceBag<D>> bag){	
		HashMap<Set<NVertex<D>>,HashMap<Connectivity<D>,Integer>> result = super.calculateBag(bag);

		//Reduce each table.
		for (Set<NVertex<D>> unusedVertices: result.keySet()){

			//Check which strategy is used to apply the reductions.
			boolean reduce = always ? (result.get(unusedVertices).size() > 1) 
									: ( result.get(unusedVertices).size() > Math.max(1, Math.pow(2, bag.data.vertices.size()-unusedVertices.size()-1)));
			
			//Apply reduction if the criteria is met.
			if (reduce){	
				//Sort the partial solution table by cost.
				ArrayList<Connectivity<D>> list = sortedConnectivityList(result.get(unusedVertices));
				
				HashMap<Connectivity<D>, Integer> table = new HashMap<Connectivity<D>, Integer>();
				ArrayList<NVertex<D>> vertexSet = new ArrayList<NVertex<D>>();
				vertexSet.addAll(bag.data.vertices);
				vertexSet.removeAll(unusedVertices);
				
				//Generate the bipartitions used to fill columns in the cut-matrix.
				//ArrayList<Connectivity<D>> biPartitions = enumerateBiPartitions(vertexSet);
				ArrayList<Connectivity<D>> biPartitions = enumerateCuts(vertexSet);
				
				
				//Fill the cut-matrix.
				int[][] matrix = new int[list.size()][biPartitions.size()];
				for (int i = 0; i < list.size(); i++){
					for (int j = 0; j < biPartitions.size(); j++){
						if (list.get(i).isRefinement(biPartitions.get(j))) {
							matrix[i][j] = 1;
							
						}
						else{
							matrix[i][j] = 0;
							
						}
					}
				}
								
				//Keep track of columns that are to be processed.
				ArrayList<Integer> swipePositions = new ArrayList<Integer>();
				for (int i = 0; i < biPartitions.size(); i++){
					swipePositions.add(i);
				}
				
				int i = 0;
				//We can stop when we have found a partial solution for each column.
				while (i < list.size() && swipePositions.size() > 0){
					//Find the leading 1.
					boolean hasOne = false;
					int swipePosition = -1;
					int j = 0;
					while(j < biPartitions.size() && !hasOne){
						if (matrix[i][j] == 1){
							hasOne = true;
							swipePosition = j;
						}
						j++;
					
					}
					
					/* Disregard the partial solution if it's row has all 0's 
					 * (linearly dependent on solutions that where found earlier). */
					if (hasOne){	
						//Add the partial solution to the result table.
						
						table.put(list.get(i), list.get(i).cost);
						
						//Swipe the column of it's leading 1.
						for (int k = i+1; k < list.size(); k++){
							if (matrix[k][swipePosition] == 1){
								//Only swipe columns that did not contain a leading 1 previously.
								for (j = 0; j < swipePositions.size(); j++){
									int column = swipePositions.get(j);
									int oldValue = matrix[k][column];
									int swipeValue = matrix[i][column];
									int newValue = (oldValue + swipeValue) % 2;
									matrix[k][column] = newValue;
									
								}						
							}
						}		
					}
					
					//Kick out the column with leading 1.
					swipePositions.remove((Integer)swipePosition);
					i++;
				}
							
				result.put(unusedVertices, table);				
			}	
		}

		return result;
	}
	
	// Sorts a table of partial solutions.
	public ArrayList<Connectivity<D>> sortedConnectivityList(HashMap<Connectivity<D>, Integer> table){
		ArrayList<Connectivity<D>> list = new ArrayList<Connectivity<D>>();
		for (Connectivity<D> connectivity: table.keySet()){
			connectivity.cost = table.get(connectivity);
			list.add(connectivity);
		}
		Collections.sort(list);
		return list;
	}
		
	// Returns an enumeration of all bipartitions of a set of vertices.
	public ArrayList<Connectivity<D>> enumerateCuts(ArrayList<NVertex<D>> list){
		ArrayList<Connectivity<D>> result = new ArrayList<Connectivity<D>>();

		//System.out.println("list: " + list.size());
		//order list
		ArrayList<NVertex<D>> orderedList = new ArrayList<NVertex<D>>();
		int k = list.size();
		for (int i = 0; i < k; i++){
			NVertex<D> smallest = list.get(0);
			int minid = list.get(0).data.id;
			for (int j = 0; j < list.size(); j++){
				if (list.get(j).data.id < minid){
					smallest = list.get(j);
					minid = list.get(j).data.id;
				}
			}
			orderedList.add(smallest);
			list.remove(smallest);
		}
		
		Set<NVertex<D>> leftBlock = new HashSet<NVertex<D>>();
		Set<NVertex<D>> rightBlock = new HashSet<NVertex<D>>();
		leftBlock.add(orderedList.get(0));
		
		ArrayList<Set<NVertex<D>>> leftBlocks = new ArrayList<Set<NVertex<D>>>();
		ArrayList<Set<NVertex<D>>> rightBlocks = new ArrayList<Set<NVertex<D>>>();
		leftBlocks.add(leftBlock);
		rightBlocks.add(rightBlock);
		
		//System.out.println("ordered: " + orderedList.size());
		for (int i = 1; i < orderedList.size(); i++){
			//System.out.println("loop");
			ArrayList<Set<NVertex<D>>> newLeftBlocks = new ArrayList<Set<NVertex<D>>>();
			ArrayList<Set<NVertex<D>>> newRightBlocks = new ArrayList<Set<NVertex<D>>>();
			
			for (int j = 0; j < leftBlocks.size(); j++){
				leftBlock = new HashSet<NVertex<D>>();
				rightBlock = new HashSet<NVertex<D>>();
				HashSet<NVertex<D>> newLeftBlock = new HashSet<NVertex<D>>();
				HashSet<NVertex<D>> newRightBlock = new HashSet<NVertex<D>>();
				
				leftBlock.addAll(leftBlocks.get(j));
				newLeftBlock.addAll(leftBlocks.get(j));
				newLeftBlock.add(orderedList.get(i));
				
				rightBlock.addAll(rightBlocks.get(j));
				newRightBlock.addAll(rightBlocks.get(j));
				newRightBlock.add(orderedList.get(i));
				
				newLeftBlocks.add(newLeftBlock);
				newLeftBlocks.add(leftBlock);
				newRightBlocks.add(rightBlock);
				newRightBlocks.add(newRightBlock);
			}
			leftBlocks = newLeftBlocks;
			rightBlocks = newRightBlocks;
		}
		
		for (int i = 0; i < leftBlocks.size(); i++){
			Connectivity<D> con = new Connectivity<D>();
			con.partition.add(leftBlocks.get(i));
			con.partition.add(rightBlocks.get(i));
			result.add(con);
		}
		//System.out.println(result.size());
		return result;
	}
	
	public ArrayList<Connectivity<D>> enumerateBiPartitions(ArrayList<NVertex<D>> list){
		ArrayList<Connectivity<D>> result = new ArrayList<Connectivity<D>>();
		ArrayList<NVertex<D>> newLeft = new ArrayList<NVertex<D>>();
		ArrayList<NVertex<D>> newRight = new ArrayList<NVertex<D>>();
		
		newLeft.add(list.get(0));
		for (int i = 1; i < list.size(); i++){
			newRight.add(list.get(i));
		}

		Connectivity<D> newConnectivity = new Connectivity<D>();
		Set<NVertex<D>> leftBlock = new HashSet<NVertex<D>>();
		Set<NVertex<D>> rightBlock = new HashSet<NVertex<D>>();
		ArrayList<ArrayList<NVertex<D>>> currentLeftGeneration = new ArrayList<ArrayList<NVertex<D>>>();
		ArrayList<ArrayList<NVertex<D>>> currentRightGeneration = new ArrayList<ArrayList<NVertex<D>>>();
		
		currentLeftGeneration.add(newLeft);
		currentRightGeneration.add(newRight);
		leftBlock.addAll(newLeft);
		rightBlock.addAll(newRight);
		newConnectivity.partition.add(leftBlock);
		newConnectivity.partition.add(rightBlock);
		result.add(newConnectivity);
		
		HashMap<NVertex<D>,Integer> priorityList = new HashMap<NVertex<D>,Integer>();
		for (int i = 0; i < list.size(); i++){
			priorityList.put(list.get(i), i);
		}
				
		for (int i = 1; i < list.size(); i++){
			ArrayList<ArrayList<NVertex<D>>> nextLeftGeneration = new ArrayList<ArrayList<NVertex<D>>>();
			ArrayList<ArrayList<NVertex<D>>> nextRightGeneration = new ArrayList<ArrayList<NVertex<D>>>();
			
			for (int j = 0; j < currentLeftGeneration.size(); j++){
				ArrayList<NVertex<D>> left = currentLeftGeneration.get(j);
				ArrayList<NVertex<D>> right = currentRightGeneration.get(j);

				int prio = priorityList.get(left.get(left.size()-1));

				for (NVertex<D> vertex: right){
					if (priorityList.get(vertex) > prio){
						newLeft = new ArrayList<NVertex<D>>();
						newRight = new ArrayList<NVertex<D>>();
						newLeft.addAll(left);
						newLeft.add(vertex);
						newRight.addAll(right);
						newRight.remove(vertex);
						nextLeftGeneration.add(newLeft);
						nextRightGeneration.add(newRight);
					}	
				}
			}
			
			for (int j = 0; j < nextLeftGeneration.size(); j++){
				leftBlock = new HashSet<NVertex<D>>();
				rightBlock = new HashSet<NVertex<D>>();
				leftBlock.addAll(nextLeftGeneration.get(j));
				rightBlock.addAll(nextRightGeneration.get(j));
				newConnectivity = new Connectivity<D>();
				newConnectivity.partition.add(leftBlock);
				newConnectivity.partition.add(rightBlock);
				result.add(newConnectivity);
			}
			
			currentLeftGeneration = nextLeftGeneration;
			currentRightGeneration = nextRightGeneration;		
		}
		//System.out.println(result.size());
		return result;
	}
	
	/* Set the condition for table-reductions.
	 * True to always apply reductions.
	 * False to apply reductions conditionally. */
	public void setCondition(boolean bool){
		always = bool;
	}
	
	

}
