/*
 * Copyright (C) 2013
 * 
 * Stefan Fafianie
 *
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * This package makes use of a selection of classes from the libtw library 
 * which is also published under the GNU Lesser General Public license.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this package; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 * 
 */

package steiner;

import java.util.BitSet;

/* Alternative representation of partial solutions using bit strings. 
 * Operators must capture the effect they have on rows in the cutmatrices
 * 
 */

public class Row implements Comparable<Row> {

	BitSet row;
	int cost;
	int size;
	
	public Row(int size){
		row = new BitSet(size);
		this.size = size;
	}
	
	public Row addVertex(int position, int total){
		Row result;
		
		//Extending a leaf solution
		if (size == 0){
			result = new Row(1);
			result.row.set(0);
			return result;
		}
		
		result = new Row(size*2);
		
		//Special case caused by asymmetry in the pattern for the first vertex
		if (position == 0){
			for (int i = 0; i < size; i++){
				if (row.get(i)){
		
					result.row.set(i);
					result.row.set(result.size-i-1);
				}
			}
		}
		//Maintain the pattern by finding indices for two new bits for every bit in the old partial solution
		else{	
			int phase = (int) Math.pow(2, total-position);
			
			for (int ind = 0; ind < result.size/2; ind++){
				int j = ind % (phase/2);
				int i = ind / (phase/2);
				if (row.get(ind)){	
					result.row.set(i*phase+j);
					result.row.set(i*phase+j+phase/2);
				}				
			}
		}
		result.cost = cost;
		return result;
	}
	
	public Row forgetVertex(int listsize, int position){
		
		//Check if the vertex is in a singleton class
		int index;
		if (position == 0){	//Asymmetrical case
			index = size-1;
		}
		else{
			index = (int)Math.pow(2, listsize-position-1);
		}
		if (row.get(index)){
			return null;
		}
		//Vertex is not in a singleton class
		else{
			Row result = new Row(size/2);
			result.cost = cost;
			
			//Asymmetrical case
			if (position == 0){
				for (int i = 0; i < size/2; i++){
					if (row.get(i) || row.get(size-i-1)){
						result.row.set(i);
					}
				}
				return result;
			}
			//Maintain pattern, find the two bits in the old partial solution to calculate the new bit
			else{
				int phase = size/(int) Math.pow(2, position-1);
				index = 0;
				for (int i = 0; i < size/phase; i++){
					for (int j = 0; j < phase/2; j++){
						if(row.get(i*phase+j) || row.get(i*phase+j+phase/2)){
							result.row.set(index);
						}
						index++;
					}
				}
				
				return result;
			}
		}		
	}
	
	//Leaf case. Does not have any cuts yet.
	public static Row leafRow(){
		Row result = new Row(0);
		result.cost = 0;
		return result;
	}
	
	//Calculate partial solution that represents an edge
	public static Row edgePrint(int total, int positionA, int positionB, int edgeWeight){
		
		//Find cuts where endpoints of the edge are contained on the same side
		Row result = new Row((int) Math.pow(2, total-1));
		int phaseA = result.size*2 / (int) Math.pow(2, positionA);
		int phaseB = result.size*2 / (int) Math.pow(2, positionB);
		for (int i = 0; i < result.size; i++){
			boolean leftA = (i%phaseA) < (phaseA/2);
			boolean leftB = (i%phaseB) < (phaseB/2);
			if ((leftA && leftB) || (!leftA && !leftB)){
				result.row.set(i);
			}
		}
		result.cost = edgeWeight;
		return result;
	}
	
	public Row joinRow(Row otherRow){
		Row result = new Row(size);
		
		//Indices in both rows correspond to the same cuts
		for (int i = 0; i < size; i++){
			if (row.get(i) && otherRow.row.get(i)){
				result.row.set(i);
			}
		}
		result.cost = cost + otherRow.cost;
		return result;
	}
	
	@Override
	public int hashCode(){
		return row.hashCode();
	}
	
	@Override
	public boolean equals(Object obj){
		if (this == obj)
	        return true;
	    if (obj == null)
	        return false;
	    if (getClass() != obj.getClass()){
	        return false;
	    }
	    else{
			final Row otherRow = (Row) obj;
	    	if (row.equals(otherRow.row))
	    		return true;
	    }
	    return false;
	}
	
	@Override
	public int compareTo(Row other) {
		return this.cost - other.cost;
	}
	
	public String toString(){
		String result = "";
		for (int i = 0; i < size; i++){
			if (row.get(i)){
				result += "1";
			}
			else{
				result += "0";
			}
		}
		return result;
	}

}
