/*
 * Copyright (C) 2013
 * 
 * Stefan Fafianie
 *
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * This package makes use of a selection of classes from the libtw library 
 * which is also published under the GNU Lesser General Public license.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this package; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 * 
 */

package steiner;

import java.util.*;

import libtw.input.GraphInput.InputData;
import libtw.ngraph.NVertex;
import libtw.ngraph.NGraph;
import libtw.ngraph.NEdge;

// Straightforward implementation of the folklore dynamic programming algorithm for Steiner-tree.
public class SteinerTreeDP<D extends InputData> {

	NGraph<NTDNiceBag<D>> niceDecomposition;
	Set<NVertex<D>> terminals;
	HashMap<Set<NVertex<D>>, Integer> edgeWeights;
	int solution;
	boolean debug;
	
	//Used for some extra statistics.
	public int totalsize = 0;
	int count;
	HashMap<Integer, Integer> maxsizetable = new HashMap<Integer, Integer>();
	HashMap<Integer, Integer> avgsizetable = new HashMap<Integer, Integer>();
	HashMap<Integer, Integer> countsizetable = new HashMap<Integer, Integer>();
	
	public void setInput(NGraph<NTDNiceBag<D>> g, Set<NVertex<D>> terminals, HashMap<Set<NVertex<D>>, Integer> edgeWeights) {
		this.niceDecomposition = g;	
		this.terminals = terminals;
		this.edgeWeights = edgeWeights;
		this.count = 0;
		this.debug = false;
	}
	
	// Start the recursive call from the root of the nice tree decomposition.
	public void run(){
		traversed = new HashSet<NVertex<NTDNiceBag<D>>>();
		NVertex<NTDNiceBag<D>> root = niceDecomposition.getVertex(0);
		count++;
		calculateBag(root);	
	}
	
	// Toggle for debugging of small instances.
	public void debug(boolean debug){
		this.debug = debug;
	}
	
	// Main recursive loop for the DP.
	Set<NVertex<NTDNiceBag<D>>> traversed;	
	public HashMap<Set<NVertex<D>>,HashMap<Connectivity<D>,Integer>> calculateBag(NVertex<NTDNiceBag<D>> bag){	
		traversed.add(bag);
		
		// Fetch the children for this bag.
		ArrayList<NVertex<NTDNiceBag<D>>> children = new ArrayList<NVertex<NTDNiceBag<D>>>();
		for (Iterator<NVertex<NTDNiceBag<D>>> neighborIterator = bag.getNeighbors(); neighborIterator.hasNext();){
			NVertex<NTDNiceBag<D>> neighbor = neighborIterator.next();
			if (!traversed.contains(neighbor)){
				children.add(neighbor);
			}
		}
		
		/* Nested hash-map structure used to find specific partial solutions in amortized O(1) time.
		 * This is used to find and eliminate duplicate partial solutions generated during the
		 * processing of forget-vertex- add-edge- and join-bags. */
		HashMap<Set<NVertex<D>>,HashMap<Connectivity<D>,Integer>> result;
		
		// Calculate the partial solution table for this bag.
		switch(bag.data.type){
		case LEAF_BAG:
			result = leafBag();
			break;
		case INTRODUCE_VERTEX_BAG:
			result = introduceVertexBag(bag.data.vertex, terminals.contains(bag.data.vertex), calculateBag(children.get(0)));
			break;
		case FORGET_VERTEX_BAG:
			result = forgetVertexBag(bag.data.vertex, calculateBag(children.get(0)));
			break;
		case INTRODUCE_EDGE_BAG:
			result = introduceEdgeBag(bag.data.edge, calculateBag(children.get(0)));
			break;
		default:
			result = joinBag(calculateBag(children.get(0)), calculateBag(children.get(1)));		
			break;
		}
		
		// Statistics.
		int size = 0;
		for (Set<NVertex<D>> key: result.keySet()){
			size += result.get(key).size();
		}
		int bagsize = bag.data.vertices.size();	
		if (maxsizetable.containsKey(bagsize)){
				int max = Math.max(size, maxsizetable.get(bagsize));
				maxsizetable.put(bagsize, max);
		}
		else{
			maxsizetable.put(bagsize, size);
		}
		if (avgsizetable.containsKey(bagsize)){
			int total = avgsizetable.get(bagsize) + size;
			avgsizetable.put(bagsize, total);
		}
		else{
			avgsizetable.put(bagsize, size);
		}
		if (countsizetable.containsKey(bagsize)){
			int count = countsizetable.get(bagsize) + 1;
			countsizetable.put(bagsize, count);
		}
		else{
			countsizetable.put(bagsize, 1);
		}
		
		if (debug){
			System.out.println(bag.data.type);
			System.out.println(count + " / " + niceDecomposition.getNumberOfVertices());
			printTable(result);
			count++;
		}
		
		return result;
	}
	
	// Use this to fetch the final solution.
	public int getSolution(){
		return this.solution;
	}
	
	// Return a table containing a single empty connectivity with cost 0.
	HashMap<Set<NVertex<D>>,HashMap<Connectivity<D>,Integer>> leafBag(){
		HashMap<Set<NVertex<D>>,HashMap<Connectivity<D>,Integer>> solutionTable = new HashMap<Set<NVertex<D>>,HashMap<Connectivity<D>,Integer>>();
		HashMap<Connectivity<D>, Integer> partitionTable = new HashMap<Connectivity<D>, Integer>();
		Connectivity<D> connectivity = new Connectivity<D>();
		partitionTable.put(connectivity, 0);
		solutionTable.put(new HashSet<NVertex<D>>(), partitionTable);
		totalsize++;
		return solutionTable;
	}
	
	// For each partial solution, add the vertex as singleton, or add it to the unused vertices.
	HashMap<Set<NVertex<D>>,HashMap<Connectivity<D>,Integer>> introduceVertexBag(
			NVertex<D> introducedVertex, 
			Boolean isTerminal, 
			HashMap<Set<NVertex<D>>,HashMap<Connectivity<D>,Integer>> childTable){
		
		HashMap<Set<NVertex<D>>,HashMap<Connectivity<D>,Integer>> solutionTable = new HashMap<Set<NVertex<D>>,HashMap<Connectivity<D>,Integer>>();
		
		// Use v: insert it as singleton in each partition.
		for (Set<NVertex<D>> unusedVertices: childTable.keySet()){
			HashMap<Connectivity<D>,Integer> childPartitionTable = childTable.get(unusedVertices);
			HashMap<Connectivity<D>,Integer> partitionTable = new HashMap<Connectivity<D>,Integer>();
			for(Connectivity<D> connectivity: childPartitionTable.keySet()){
				int cost = childPartitionTable.get(connectivity);
				Connectivity<D> newConnectivity = connectivity.addVertex(introducedVertex);
				totalsize++;
				partitionTable.put(newConnectivity, cost);
			}
			solutionTable.put(unusedVertices, partitionTable);			
		}
			
		// Don't use v: insert it in the unused parts.
		if (!isTerminal){
			for(Set<NVertex<D>> unusedVertices: childTable.keySet()){
				HashMap<Connectivity<D>,Integer> childPartitionTable = childTable.get(unusedVertices);
				Set<NVertex<D>> newUnusedVertices = new HashSet<NVertex<D>>();
				newUnusedVertices.addAll(unusedVertices);
				newUnusedVertices.add(introducedVertex);
				//totalsize += childPartitionTable.size();
				solutionTable.put(newUnusedVertices, childPartitionTable);
			}
		}
		
		return solutionTable;
	}
	
	/* Remove the vertex from partial solutions in the table. In the easy case the vertex wasn't used.
	 * In the other case we remove it if it is not a singleton in the partial solution and eliminate it
	 * otherwise. Duplicate partial solutions can be generated during this process. The most expensive
	 * ones are eliminated. */
	HashMap<Set<NVertex<D>>,HashMap<Connectivity<D>,Integer>> forgetVertexBag(
			NVertex<D> forgottenVertex, 
			HashMap<Set<NVertex<D>>,HashMap<Connectivity<D>,Integer>> childTable){
		
		HashMap<Set<NVertex<D>>,HashMap<Connectivity<D>,Integer>> solutionTable = new HashMap<Set<NVertex<D>>,HashMap<Connectivity<D>,Integer>>();
		
		Set<Set<NVertex<D>>> usedForgottenVertex = new HashSet<Set<NVertex<D>>>();
		Set<Set<NVertex<D>>> unusedForgottenVertex = new HashSet<Set<NVertex<D>>>();
		
		for (Set<NVertex<D>> unusedVertices: childTable.keySet()){
			if (unusedVertices.contains(forgottenVertex)){
				unusedForgottenVertex.add(unusedVertices);
			}
			else{
				usedForgottenVertex.add(unusedVertices);
			}
		}
					
		// Commit to using v.
		for (Set<NVertex<D>> unusedVertices: usedForgottenVertex){
			HashMap<Connectivity<D>,Integer> childPartitionTable = childTable.get(unusedVertices);
			HashMap<Connectivity<D>,Integer> partitionTable = new HashMap<Connectivity<D>, Integer>();
			for(Connectivity<D> connectivity: childPartitionTable.keySet()){			
				int cost = childPartitionTable.get(connectivity);	
				Connectivity<D> newConnectivity = connectivity.forgetVertex(forgottenVertex, unusedVertices.size()==0);
				
				if (newConnectivity != null){	
					totalsize++;
					if (partitionTable.containsKey(newConnectivity)){
						cost = Math.min(cost, partitionTable.get(newConnectivity));
					}
					partitionTable.put(newConnectivity, cost);
					if (unusedVertices.size()==0){
						// This happens in the root bag only -> final solution.
						solution = cost;
					}
				}
			}
			solutionTable.put(unusedVertices, partitionTable);	
		}
		
		// Commit to not using v.
		for (Set<NVertex<D>> unusedVertices: unusedForgottenVertex){			
			HashMap<Connectivity<D>,Integer> childPartitionTable = childTable.get(unusedVertices);
			Set<NVertex<D>> newUnusedVertices = new HashSet<NVertex<D>>();
			newUnusedVertices.addAll(unusedVertices);
			newUnusedVertices.remove(forgottenVertex);	
			if (solutionTable.keySet().contains(newUnusedVertices)){
				HashMap<Connectivity<D>,Integer> previousPartitionTable = solutionTable.get(newUnusedVertices);
				for(Connectivity<D> connectivity: childPartitionTable.keySet()){
					//totalsize++;
					int cost = childPartitionTable.get(connectivity);
					if (previousPartitionTable.containsKey(connectivity)){
						cost = Math.min(cost, previousPartitionTable.get(connectivity));
					}
					previousPartitionTable.put(connectivity, cost);						
				}
			}
			else{
				solutionTable.put(newUnusedVertices, childPartitionTable);
			}			
		}
		
		return solutionTable;
	}
	
	/* Introduce an edge to the partial solutions. For each solution we consider adding or not adding the edge.
	 * When we add the edge we have to make sure that both incident vertices where used in the connectivity. 
	 * Again, duplicate solutions may be generated and are removed. */
	HashMap<Set<NVertex<D>>,HashMap<Connectivity<D>,Integer>> introduceEdgeBag(
			NEdge<D> introducedEdge, 
			HashMap<Set<NVertex<D>>,HashMap<Connectivity<D>,Integer>> childTable){

		HashMap<Set<NVertex<D>>,HashMap<Connectivity<D>,Integer>> solutionTable = new HashMap<Set<NVertex<D>>,HashMap<Connectivity<D>,Integer>>();
				
		//Don't use the edge
		for(Set<NVertex<D>> unusedVertices: childTable.keySet()){
			HashMap<Connectivity<D>,Integer> childPartitionTable = childTable.get(unusedVertices);
			HashMap<Connectivity<D>,Integer> partitionTable = new HashMap<Connectivity<D>,Integer>();
			for(Connectivity<D> connectivity: childPartitionTable.keySet()){
				//totalsize++;
				partitionTable.put(connectivity, childPartitionTable.get(connectivity));
			}
			solutionTable.put(unusedVertices, partitionTable);
		}
		
		//Use the edge if it improves anything
		for(Set<NVertex<D>> unusedVertices: childTable.keySet()){		
			if(!unusedVertices.contains(introducedEdge.a) && !unusedVertices.contains(introducedEdge.b)){
				HashMap<Connectivity<D>,Integer> childPartitionTable = childTable.get(unusedVertices);
				HashMap<Connectivity<D>,Integer> partitionTable = solutionTable.get(unusedVertices);
				//glue partitions
				for(Connectivity<D> connectivity: childPartitionTable.keySet()){	
					totalsize++;
					Set<NVertex<D>> edge = new HashSet<NVertex<D>>();
					edge.add(introducedEdge.a);
					edge.add(introducedEdge.b);
					int cost = childPartitionTable.get(connectivity) + edgeWeights.get(edge);
					Connectivity<D> newConnectivity = connectivity.addEdge(introducedEdge);
					if (partitionTable.containsKey(newConnectivity)){
						cost = Math.min(cost, partitionTable.get(newConnectivity));
					}
					partitionTable.put(newConnectivity, cost);
				}
			}	
		}

		return solutionTable;
	}
		
	/* Combine all combinations of partial solutions that use the same base set of vertices.  
	 * Again, duplicate solutions may be generated and are removed. */
	HashMap<Set<NVertex<D>>,HashMap<Connectivity<D>,Integer>> joinBag(	
			HashMap<Set<NVertex<D>>,HashMap<Connectivity<D>,Integer>> leftChildTable,
			HashMap<Set<NVertex<D>>,HashMap<Connectivity<D>,Integer>> rightChildTable){

		HashMap<Set<NVertex<D>>,HashMap<Connectivity<D>,Integer>> solutionTable = new HashMap<Set<NVertex<D>>,HashMap<Connectivity<D>,Integer>>();

		Set<Set<NVertex<D>>> jointKeySet = new HashSet<Set<NVertex<D>>>();
		jointKeySet.addAll(leftChildTable.keySet());
		jointKeySet.addAll(rightChildTable.keySet());
				
		for(Set<NVertex<D>> unusedVertices : jointKeySet){
			HashMap<Connectivity<D>, Integer> leftChildPartitionTable = leftChildTable.get(unusedVertices);
			HashMap<Connectivity<D>, Integer> rightChildPartitionTable = rightChildTable.get(unusedVertices);
			HashMap<Connectivity<D>, Integer> partitionTable = new HashMap<Connectivity<D>, Integer>();
					
			for (Connectivity<D> leftConnectivity: leftChildPartitionTable.keySet()){
				int leftCost = leftChildPartitionTable.get(leftConnectivity);
				for(Connectivity<D> rightConnectivity: rightChildPartitionTable.keySet()){
					int rightCost = rightChildPartitionTable.get(rightConnectivity);
					int totalCost = leftCost + rightCost;
									
					//Combine partitions. 
					Connectivity<D> newConnectivity = leftConnectivity.joinConnectivity(rightConnectivity);
					totalsize++;
					if (partitionTable.containsKey(newConnectivity)){
						totalCost = Math.min(totalCost,  partitionTable.get(newConnectivity));
					}
					partitionTable.put(newConnectivity, totalCost);
				}					
			}		
			solutionTable.put(unusedVertices, partitionTable);	
		}
		return solutionTable;	
	}
				
	// Prints the tables during debugging. Using small problem instances is advised. 
	public void printTable(HashMap<Set<NVertex<D>>,HashMap<Connectivity<D>,Integer>> table){
		System.out.println("------------");
		for (Set<NVertex<D>> key:table.keySet()){
			HashMap<Connectivity<D>,Integer> sub_solution_table = table.get(key);
			String unused = "{ ";
			for(NVertex<D> vertex: key){
				unused+= vertex.data.name + " ";
			}
			unused += "}" + " | ";
			for (Connectivity<D> sub_key: sub_solution_table.keySet()){
				
				int solution = sub_solution_table.get(sub_key);
				
				String partition = "{ ";
				for(Set<NVertex<D>> vertexset: sub_key.partition){
					partition += "{ ";
					for(NVertex<D> vertex: vertexset){
						partition+= vertex.data.name + " ";
					}
					partition += "} ";
							
				}
				partition += "} | ";
				
				
				System.out.println(unused + partition + solution);
			}
			if (sub_solution_table.isEmpty()){
				System.out.println(unused += "*");
			}
		}
		System.out.println("");
	}
	
	
}
